@@ -77,6 +77,87 @@ class AADictionaryTestCase(unittest.TestCase):
         self.assertRaises(LG_Error, Dictionary, dummy_lang + '2')
         self.assertIn(dummy_lang + '2', save_stderr.divert_end())
 
+# Check absolute and relative dictionary access.
+# Check also that the dictionary language is set correctly.
+#
+# We suppose here that this test program is located somewhere in the source
+# directory, 1 to 4 levels under it, that the data directory is named 'data',
+# and that it has a parallel directory called 'link-grammar'.
+DATA_DIR = 'data'
+PARALLEL_DIR = 'link-grammar'
+class ABDictionaryLocationTestCase(unittest.TestCase):
+    abs_datadir = None
+
+    @classmethod
+    def setUpClass(cls):
+        cls.po = ParseOptions(verbosity=0)
+        cls.original_directory = os.getcwd()
+
+        # Find the 'data' directory in the source directory.
+        os.chdir(clg.test_data_srcdir)
+        up = ''
+        for _ in range(1, 4):
+            up = '../' + up
+            datadir = up + DATA_DIR
+            if os.path.isdir(datadir):
+                break
+            datadir = ''
+
+        if not datadir:
+            assert False, 'Cannot find source directory dictionary data'
+        cls.abs_datadir = os.path.abspath(datadir)
+
+    @classmethod
+    def tearDownClass(cls):
+        del cls.po
+        os.chdir(cls.original_directory)
+
+    def test_open_absolute_path(self):
+        d = Dictionary(self.abs_datadir + '/en')
+        self.assertEqual(str(d), 'en')
+        if os.name == 'nt':
+            d = Dictionary(self.abs_datadir + r'\en')
+            self.assertEqual(str(d), 'en')
+
+    def test_open_relative_path_from_data_directory(self):
+        os.chdir(self.abs_datadir)
+        d = Dictionary('./en')
+        self.assertEqual(str(d), 'en')
+        if os.name == 'nt':
+            d = Dictionary(r'.\en')
+            self.assertEqual(str(d), 'en')
+
+    def test_open_lang_from_data_directory(self):
+        os.chdir(self.abs_datadir)
+        d = Dictionary('en')
+        self.assertEqual(str(d), 'en')
+
+    # Test use of the internal '..' path
+    def test_open_from_a_language_directory(self):
+        os.chdir(self.abs_datadir + '/ru')
+        d = Dictionary('en')
+        self.assertEqual(str(d), 'en')
+
+    def test_open_relative_path_from_data_parent_directory(self):
+        os.chdir(self.abs_datadir + '/..')
+        d = Dictionary('data/en')
+        self.assertEqual(str(d), 'en')
+        if os.name == 'nt':
+            d = Dictionary(r'data\en')
+            self.assertEqual(str(d), 'en')
+
+    # Test use of the internal './data' path.
+    def test_open_from_data_parent_directory(self):
+        os.chdir(self.abs_datadir + '/..')
+        d = Dictionary('en')
+        self.assertEqual(str(d), 'en')
+
+    # Test use of the internal '../data' path.
+    def test_open_from_a_parallel_directory(self):
+        os.chdir(self.abs_datadir + '/../' + PARALLEL_DIR)
+        d = Dictionary('en')
+        self.assertEqual(str(d), 'en')
+
 class BParseOptionsTestCase(unittest.TestCase):
     def test_setting_verbosity(self):
         po = ParseOptions()
@@ -322,33 +403,11 @@ class DBasicParsingTestCase(unittest.TestCase):
         linkage = self.parse_sent("This is a silly sentence.")[0]
         self.assertEqual([len(l) for l in linkage.links()], [6,2,1,1,3,2,1,1,1])
 
-    def test_dictionary_locale_definition(self):
-        if is_python2(): # Locale stuff seems to be broken
-            raise unittest.SkipTest("Test not supported with Python2")
-
-        # python2: Gets system locale (getlocale() is not better)
-        oldlocale = locale.setlocale(locale.LC_CTYPE, None)
-        #print('Current locale:', oldlocale)
-        #print('toupper hij:', 'hij'.upper())
-
-        tr_locale = 'tr_TR.UTF-8' if os.name != 'nt' else 'Turkish'
-        locale.setlocale(locale.LC_CTYPE, tr_locale)
-        #print('Turkish locale:', locale.setlocale(locale.LC_CTYPE, None))
-
-        # python2: prints HiJ (lowercase small i in the middle)
-        #print('toupper hij:', 'hij'.upper())
-
-        self.assertEqual(list(self.parse_sent('Is it fine?')[0].words()),
-                         ['LEFT-WALL', 'is.v', 'it', 'fine.a', '?', 'RIGHT-WALL'])
-
-        locale.setlocale(locale.LC_CTYPE, oldlocale)
-        #print("Restored locale:", locale.setlocale(locale.LC_CTYPE))
-        #print('toupper hij:', 'hij'.upper())
-
     # If \w is supported, other \ shortcuts are hopefully supported too.
     def test_regex_class_shortcut_support(self):
         r"""Test that regexes support \w"""
-        linkage = self.parse_sent("This is a _regex_ive regex test")[0]
+        po = ParseOptions(display_morphology=False)
+        linkage = self.parse_sent("This is a _regex_ive regex test", po)[0]
         self.assertEqual(linkage.word(4), '_regex_ive[!].a')
 
     def test_timer_exhausted_exception(self):
@@ -552,7 +611,7 @@ class FSATsolverTestCase(unittest.TestCase):
 class HEnglishLinkageTestCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
-        cls.d, cls.po = Dictionary(), ParseOptions(linkage_limit=300)
+        cls.d, cls.po = Dictionary(), ParseOptions(linkage_limit=300, display_morphology=False)
 
     @classmethod
     def tearDownClass(cls):
@@ -689,7 +748,7 @@ class HEnglishLinkageTestCase(unittest.TestCase):
     # Some parses are fractionally preferred over others...
     def test_g_fractions(self):
         self.assertEqual(list(self.parse_sent('A player who is injured has to leave the field')[0].words()),
-             ['LEFT-WALL', 'a', 'player.n', 'who', 'is.v', 'injured.a', 'has.v', 'to.r', 'leave.v', 'the', 'field.n', 'RIGHT-WALL'])
+             ['LEFT-WALL', 'a', 'player.n', 'who', 'is.v', 'injured.v-d', 'has.v', 'to.r', 'leave.v', 'the', 'field.n', 'RIGHT-WALL'])
 
         self.assertEqual(list(self.parse_sent('They ate a special curry which was recommended by the restaurant\'s owner')[0].words()),
              ['LEFT-WALL', 'they', 'ate.v-d', 'a', 'special.a', 'curry.s',
@@ -733,6 +792,10 @@ class HEnglishLinkageTestCase(unittest.TestCase):
 class GSQLDictTestCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
+        if os.name == 'nt' and \
+                -1 == clg.linkgrammar_get_configuration().lower().find('mingw'):
+            raise unittest.SkipTest("No SQL dict support yet on the MSVC build")
+
         #clg.parse_options_set_verbosity(clg.parse_options_create(), 3)
         cls.d, cls.po = Dictionary(lang='demo-sql'), ParseOptions()
 
@@ -824,6 +887,68 @@ class ZENLangTestCase(unittest.TestCase):
         self.assertEqual(len(linkages), 2)
         self.assertEqual(linkages.next().unused_word_cost(), 1)
 
+class JADictionaryLocaleTestCase(unittest.TestCase):
+    @classmethod
+    def setUpClass(cls):
+        if is_python2(): # Locale stuff seems to be broken
+            raise unittest.SkipTest("Test not supported with Python2")
+
+        # python2: Gets system locale (getlocale() is not better)
+        cls.oldlocale = locale.setlocale(locale.LC_CTYPE, None)
+        #print('Current locale:', oldlocale)
+        #print('toupper hij:', 'hij'.upper())
+
+        tr_locale = 'tr_TR.UTF-8' if os.name != 'nt' else 'Turkish'
+        try:
+            locale.setlocale(locale.LC_CTYPE, tr_locale)
+        except locale.Error as e: # Most probably tr_TR.UTF-8 is not installed
+            raise unittest.SkipTest("Locale {}: {}".format(tr_locale, e))
+
+        #print('Turkish locale:', locale.setlocale(locale.LC_CTYPE, None))
+        # python2: prints HiJ (lowercase small i in the middle)
+        #print('toupper hij:', 'hij'.upper())
+
+        cls.d, cls.po = Dictionary(lang='en'), ParseOptions()
+
+    @classmethod
+    def tearDownClass(cls):
+        locale.setlocale(locale.LC_CTYPE, cls.oldlocale)
+        #print("Restored locale:", locale.setlocale(locale.LC_CTYPE))
+        #print('toupper hij:', 'hij'.upper())
+        del cls.d, cls.po, cls.oldlocale
+
+    def test_dictionary_locale_definition(self):
+        linkage = Sentence('Is it fine?', self.d, self.po).parse().next()
+        self.assertEqual(list(linkage.words())[1], 'is.v')
+
+# FIXME: Use a special testing dictionary for checks like that.
+class JBDictCostReadingTestCase(unittest.TestCase):
+    @classmethod
+    def setUpClass(cls):
+        if is_python2(): # Locale stuff seems to be broken
+            raise unittest.SkipTest("Test not supported with Python2")
+
+        cls.oldlocale = locale.setlocale(locale.LC_CTYPE, None)
+        ru_locale = 'ru_RU.UTF-8' if os.name != 'nt' else 'Russian'
+        try:
+            locale.setlocale(locale.LC_NUMERIC, ru_locale)
+        except locale.Error as e: # Most probably ru_RU.UTF-8 is not installed
+            del cls.oldlocale
+            raise unittest.SkipTest("Locale {}: {}".format(ru_locale, e))
+        # The dict read must be performed after the locale change.
+        cls.d, cls.po = Dictionary(lang='en'), ParseOptions()
+
+    @classmethod
+    def tearDownClass(cls):
+        locale.setlocale(locale.LC_CTYPE, cls.oldlocale)
+        del cls.d, cls.po, cls.oldlocale
+
+    # When a comma-separator LC_NUMERIC affects the dict cost conversion,
+    # the 4th word is 'white.v'.
+    def test_cost_sensitive_parse(self):
+        linkage = Sentence('Is the bed white?', self.d, self.po).parse().next()
+        self.assertEqual(list(linkage.words())[4], 'white.a')
+
 class ZENConstituentsCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
@@ -902,16 +1027,19 @@ class ZRULangTestCase(unittest.TestCase):
         return list(Sentence(text, self.d, self.po).parse())
 
     def test_a_getting_num_of_words(self):
+        self.po.display_morphology = False
         #Words include punctuation and a 'LEFT-WALL' and 'RIGHT_WALL'
         self.assertEqual(self.parse_sent('это тести.')[0].num_of_words(), 5)
         self.assertEqual(self.parse_sent('вверху плыли редкие облачка.')[0].num_of_words(), 7)
 
     def test_b_getting_words(self):
+        self.po.display_morphology = False
         self.assertEqual(list(self.parse_sent('вверху плыли редкие облачка.')[0].words()),
             ['LEFT-WALL', 'вверху.e', 'плыли.vnndpp', 'редкие.api',
                 'облачка.ndnpi', '.', 'RIGHT-WALL'])
 
     def test_c_getting_links(self):
+        self.po.display_morphology = False
         sent = 'вверху плыли редкие облачка.'
         linkage = self.parse_sent(sent)[0]
         self.assertEqual(linkage.link(0),
@@ -1031,7 +1159,7 @@ def linkage_testfile(self, lgdict, popt, desc = ''):
 
         # Lines starting with "-" contain a Parse Option
         elif line[0] == '-':
-            exec('popt.' + line[1:], None, locals())
+                exec('popt.' + line[1:]) in {}, locals()
 
         elif line[0] in '%\r\n':
             pass
