@@ -119,9 +119,6 @@ Parse_Options parse_options_create(void)
 	po->max_null_count = 0;
 	po->islands_ok = false;
 	po->use_sat_solver = false;
-#ifdef USE_VITERBI
-	po->use_viterbi = false;
-#endif
 	po->linkage_limit = 100;
 #if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 	po->use_spell_guess = 7;
@@ -147,7 +144,7 @@ Parse_Options parse_options_create(void)
 	po->repeatable_rand = true;
 	po->resources = resources_create();
 	po->use_cluster_disjuncts = false;
-	po->display_morphology = false;
+	po->display_morphology = true;
 
 	return po;
 }
@@ -294,16 +291,6 @@ bool parse_options_get_use_sat_parser(Parse_Options opts) {
 	return opts->use_sat_solver;
 }
 
-#ifdef USE_VITERBI
-void parse_options_set_use_viterbi(Parse_Options opts, bool dummy) {
-	opts->use_viterbi = dummy;
-}
-
-bool parse_options_get_use_viterbi(Parse_Options opts) {
-	return opts->use_viterbi;
-}
-#endif
-
 void parse_options_set_linkage_limit(Parse_Options opts, int dummy)
 {
 	opts->linkage_limit = dummy;
@@ -455,7 +442,6 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 	sent->dict = dict;
 	sent->string_set = string_set_create();
 	sent->rand_state = global_rand_state;
-	sent->disjuncts_connectors_memblock = NULL;
 
 	sent->postprocessor = post_process_new(dict->base_knowledge);
 
@@ -530,6 +516,7 @@ void sentence_delete(Sentence sent)
 	free_sentence_words(sent);
 	wordgraph_delete(sent);
 	string_set_delete(sent->string_set);
+	string_id_delete(sent->connector_suffix_id);
 	free_linkages(sent);
 	post_process_free(sent->postprocessor);
 	post_process_free(sent->constituent_pp);
@@ -586,7 +573,6 @@ double sentence_disjunct_cost(Sentence sent, LinkageIdx i)
 {
 	if (!sent) return 0.0;
 
-	/* The sat solver (currently) fails to fill in link_info */
 	if (!sent->lnkages) return 0.0;
 	if (sent->num_linkages_alloced <= i) return 0.0; /* bounds check */
 	return sent->lnkages[i].lifo.disjunct_cost;
@@ -596,7 +582,6 @@ int sentence_link_cost(Sentence sent, LinkageIdx i)
 {
 	if (!sent) return 0;
 
-	/* The sat solver (currently) fails to fill in link_info */
 	if (!sent->lnkages) return 0;
 	if (sent->num_linkages_alloced <= i) return 0; /* bounds check */
 	return sent->lnkages[i].lifo.link_cost;
@@ -671,10 +656,6 @@ int sentence_parse(Sentence sent, Parse_Options opts)
  * Definitions for linkgrammar_get_configuration().
  */
 
-/* __VA_ARGS__ must be used because arguments may contain commas. */
-#define lg_xstr(...) lg_str(__VA_ARGS__)
-#define lg_str(...) #__VA_ARGS__
-
 #ifdef __STDC_VERSION__
 #define LG_S1 "__STDC_VERSION__=" lg_xstr(__STDC_VERSION__)
 #else
