@@ -15,13 +15,13 @@
 #include <sys/types.h>
 #include <sys/stat.h>                   // fstat()
 
-#ifndef _WIN32
+#ifndef _MSC_VER
 	#include <unistd.h>
 #else
 	#include <windows.h>
 	#include <Shlwapi.h>                 // PathRemoveFileSpecA()
 	#include <direct.h>                  // getcwd()
-#endif /* _WIN32 */
+#endif /* _MSC_VER */
 
 #include <stdlib.h>
 #include <string.h>
@@ -37,11 +37,8 @@
 	#define DIR_SEPARATOR "/"
 #endif /*_WIN32 */
 
-#define IS_DIR_SEPARATOR(ch) (DIR_SEPARATOR[0] == (ch))
-#if !defined(DICTIONARY_DIR) || defined(__MINGW32__)
-	#define DEFAULTPATH NULL
-#else
-	#define DEFAULTPATH DICTIONARY_DIR
+#ifndef DICTIONARY_DIR
+#define DICTIONARY_DIR NULL
 #endif
 
 /* =========================================================== */
@@ -50,6 +47,23 @@
 #define MAX_PATH_NAME 200     /* file names (including paths)
                                  should not be longer than this */
 
+/**
+ * Find the location of the last directory separator (Windows/POSIX).
+ * Return NULL if none found.
+ * It doesn't modify its argument, but const is not used because
+ * it is to be called also with non-const argument.
+ */
+char *find_last_dir_separator(char *path)
+{
+	char *dirsep = NULL;
+	size_t pathlen = strlen(path);
+
+	for (size_t p = pathlen; p > 0; p--)
+		if (('/' == path[p]) || ('\\' == path[p])) return &path[p];
+
+	return dirsep;
+}
+
 char * join_path(const char * prefix, const char * suffix)
 {
 	char * path;
@@ -64,9 +78,9 @@ char * join_path(const char * prefix, const char * suffix)
 	 * only if the prefix isn't already terminated by a path sep.
 	 */
 	prel = strlen(path);
-	if (0 < prel && path[prel-1] != DIR_SEPARATOR[0])
+	if (0 < prel && (path[prel-1] != '/') && (path[prel-1] != '\\'))
 	{
-		path[prel] = DIR_SEPARATOR[0];
+		path[prel] = '/';
 		path[prel+1] = '\0';
 	}
 	strcat(path, suffix);
@@ -82,8 +96,7 @@ char * join_path(const char * prefix, const char * suffix)
  */
 static char * custom_data_dir = NULL;
 
-static void free_custom_data_dir(void)
-{
+static void free_custom_data_dir(void) {
 	free(custom_data_dir);
 }
 
@@ -105,44 +118,164 @@ char * dictionary_get_data_dir(void)
 		return data_dir;
 	}
 
-#ifdef _WIN32
-	/* Dynamically locate invocation directory of our program.
-	 * Non-ASCII characters are not supported (files will not be found). */
-	char prog_path[MAX_PATH_NAME];
+	return NULL;
+}
 
-	if (!GetModuleFileNameA(NULL, prog_path, sizeof(prog_path)))
-	{
-		prt_error("Warning: GetModuleFileName error %d\n", (int)GetLastError());
+#ifdef _MSC_VER
+static const char *get_dictionary_dir(bool);
+static void free_dictionary_dir(void)
+{
+	get_dictionary_dir(false);
+}
+#endif // _MSC_VER
+
+/**
+ * Return the path of the system dictionary directory.
+ * If DICTIONARY_DIR is an absolute path (must be so on systems other than
+ * Windows) then it is directly used.
+ * On Windows (currently implemented on MSVC only), a relative
+ * DICTIONARY_DIR is combined with the location of this library DLL, and
+ * if it includes a drive letter or is a UNC path then it must be
+ * absolute. It shouldn't end with a directory separator.
+ * @param find If true, find the system dictionary directory; else free it
+ * if needed.
+ * @return The actual dictionary directory.
+ */
+static const char *get_dictionary_dir(bool find)
+{
+#ifndef _MSC_VER
+		return DICTIONARY_DIR;
+#else
+	/* We are on Windows. DICTIONARY_DIR, if defined, can be relative or
+	 * absolute. An absolute path must start with a drive letter or be a
+	 * UNC path, and a relative path should not start with a drive letter
+	 * or be a UNC path.
+	 * In case it is not so, the result is undefined.  If it is a relative
+	 * path, it is concatenated to the program location. */
+	static const char *dictionary_dir;
+
+	if (!find) {
+		free((void *)dictionary_dir);
+		return NULL;
 	}
-	else
+
+	/* If we already found it, just return it. */
+	if (NULL != dictionary_dir) return dictionary_dir;
+
+	dictionary_dir = DICTIONARY_DIR;
+	/* If it is NULL or "", assume it is relative to the current directory. */
+	if ((NULL == dictionary_dir) || ('\0' == dictionary_dir[1]))
+		dictionary_dir = ".";
+
+	/* DICTIONARY_DIR is already an absolute path - return it.
+	 * Note: UNC paths are supposed to be absolute. */
+	if (0 == strncmp(dictionary_dir, "\\\\", 2)) return DICTIONARY_DIR;
+	if (0 == strncmp(dictionary_dir, "//", 2)) return DICTIONARY_DIR;
+
+	/* It includes a drive letter - suppose it's absolute so just return it. */
+	if ((strlen(dictionary_dir) > 2) && (':' == dictionary_dir[1]))
+		return DICTIONARY_DIR;
+
+	/* Here DICTIONARY_DIR is a relative path (to be concatenated to the
+	 * location of this library DLL), or an absolute path w/o a drive
+	 * letter (to be prepended with the drive letter of our program).
+	 * Non-ASCII characters in the DLL location are not supported (files
+	 * will not be found).
+	 * Note: In case of an error, we return dictionary_dir.
+	 * Is there anything better to do? */
+	char dll_path[MAX_PATH_NAME] = "";
+	HMODULE dll_hm = NULL;
+
+	/* First find the module handle of this library DLL. */
+	if (!GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
+	                       GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
+	                       (LPCSTR) &get_dictionary_dir, &dll_hm))
 	{
-		if (NULL == prog_path)
+		prt_error("Warning: GetModuleHandleEx error %d\n", (int)GetLastError());
+		return dictionary_dir; /* Is there anything better to do? */
+	}
+
+	/* Then use it to find the DLL path. */
+	if (!GetModuleFileNameA(dll_hm, dll_path, sizeof(dll_path)))
+	{
+		prt_error("Warning: GetModuleFileNameA error %d\n", (int)GetLastError());
+		return dictionary_dir;
+	}
+
+	if ('\0' == dll_path[0])
+	{
+		/* Can it happen? */
+		prt_error("Warning: GetModuleFileNameA didn't return a path!\n");
+		return dictionary_dir;
+	}
+
+	if (!PathRemoveFileSpecA(dll_path))
+	{
+		prt_error("Warning: Cannot get directory from LG DLL path '%s'!\n",
+					 dll_path);
+		return dictionary_dir;
+	}
+
+	/* Unconvertible characters are marked as '?' */
+	if (NULL != strchr(dll_path, '?'))
+	{
+		prt_error("Warning: Directory of LG DLL (%s) "
+		          "contains unsupported characters\n", dll_path);
+		return dictionary_dir;
+	}
+
+	/* Just a sanity check of the directory length. */
+	if (strlen(dll_path) < 3)
+	{
+		prt_error("Warning: DLL directory name '%s' too short!\n", dll_path);
+		return dictionary_dir;
+	}
+
+	lgdebug(D_USER_FILES, "Debug: Directory of LG DLL: %s\n", dll_path);
+
+	char *combined_dictionary_dir;
+
+	if (('\\' == dictionary_dir[0]) || ('/' == dictionary_dir[0]))
+	{
+		/* DICTIONARY_DIR is an absolute path without a drive
+		 * (UNC paths already returned unmodified, above).
+		 * Prepend the drive or the host of our program location. */
+		size_t prefix_len = 0;
+
+		if (dll_path[1] == ':')
 		{
-			/* Can it happen? */
-			prt_error("Warning: GetModuleFileName returned a NULL program path!\n");
+			/* "X:path" */
+			prefix_len = 2;
 		}
 		else
 		{
-			if (!PathRemoveFileSpecA(prog_path))
-			{
-				prt_error("Warning: Cannot get directory from program path '%s'!\n",
-				          prog_path);
-			}
-			else
-			{
-				/* Unconvertible characters are marked as '?' */
-				const char *unsupported = (NULL != strchr(prog_path, '?')) ?
-					" (containing unsupported character)" : "";
-
-				lgdebug(D_USER_FILES, "Debug: Directory of executable: %s%s\n",
-				        unsupported, prog_path);
-				data_dir = safe_strdup(prog_path);
-			}
+			/* \\host\path */
+			const char *hostend = strchr(dll_path+3, '\\');
+			if (NULL == hostend)
+				hostend = strchr(dll_path+3, '/');
+			if (NULL != hostend)
+				prefix_len = (size_t)(hostend - dll_path);
 		}
+		size_t len = prefix_len + strlen(dictionary_dir) + 1;
+		combined_dictionary_dir = malloc(len);
+		strncpy(combined_dictionary_dir, dll_path, prefix_len);
+		strcpy(combined_dictionary_dir + prefix_len, dictionary_dir);
+	}
+	else
+	{
+		size_t len = strlen(dll_path)+1+strlen(dictionary_dir)+1;
+		combined_dictionary_dir = malloc(len);
+		strcpy(combined_dictionary_dir, dll_path);
+		strcat(combined_dictionary_dir, "\\"); /* Mixing / and \ is allowed. */
+		strcat(combined_dictionary_dir, dictionary_dir);
 	}
-#endif /* _WIN32 */
 
-	return data_dir;
+	dictionary_dir = combined_dictionary_dir;
+	atexit(free_dictionary_dir);
+	lgdebug(D_USER_FILES, "Debug: Using dictionary directory '%s'\n",
+	        dictionary_dir);
+	return dictionary_dir;
+#endif /* _MSC_VER */
 }
 
 static void *dict_file_open(const char *fullname, const void *how)
@@ -184,6 +317,7 @@ void * object_open(const char *filename,
 	char *completename = NULL;
 	void *fp = NULL;
 	char *data_dir = NULL;
+	const char *dictionary_dir = NULL;
 	const char **path = NULL;
 
 	if (NULL == filename)
@@ -197,14 +331,17 @@ void * object_open(const char *filename,
 
 	if (NULL == path_found)
 	{
+		dictionary_dir = get_dictionary_dir(true);
 		data_dir = dictionary_get_data_dir();
 		if (verbosity_level(D_USER_FILES))
 		{
 			char cwd[MAX_PATH_NAME];
 			char *cwdp = getcwd(cwd, sizeof(cwd));
 			prt_error("Debug: Current directory: %s\n", NULL == cwdp ? "NULL": cwdp);
-			prt_error("Debug: Last-resort data directory: %s\n",
+			prt_error("Debug: Data directory: %s\n",
 					  data_dir ? data_dir : "NULL");
+			prt_error("Debug: System data directory: %s\n",
+					  dictionary_dir ? dictionary_dir : "NULL");
 		}
 	}
 
@@ -231,11 +368,11 @@ void * object_open(const char *filename,
 		{
 			path_found,
 			".",
-			"." DIR_SEPARATOR "data",
+			"./data",
 			"..",
-			".." DIR_SEPARATOR "data",
+			"../data",
 			data_dir,
-			DEFAULTPATH,
+			dictionary_dir,
 		};
 		size_t i = sizeof(dictpath)/sizeof(dictpath[0]);
 
@@ -263,10 +400,11 @@ void * object_open(const char *filename,
 			prt_error("Info: Dictionary found at %s\n", pfnd);
 		for (size_t i = 0; i < 2; i++)
 		{
-			char *root = strrchr(pfnd, DIR_SEPARATOR[0]);
+			char *root = find_last_dir_separator(pfnd);
 			if (NULL != root) *root = '\0';
 		}
 		path_found = pfnd;
+		lgdebug(D_USER_FILES, "Debug: Using dictionary path \"%s\"\n", path_found);
 	}
 
 	free(data_dir);
