@@ -112,7 +112,7 @@ dictionary_six_str(const char * lang,
 
 	/* Language and file-name stuff */
 	dict->string_set = string_set_create();
-	t = strrchr (lang, '/');
+	t = find_last_dir_separator((char *)lang);
 	t = (NULL == t) ? lang : t+1;
 	dict->lang = string_set_add(t, dict->string_set);
 	lgdebug(D_USER_FILES, "Debug: Language: %s\n", dict->lang);
@@ -154,6 +154,11 @@ dictionary_six_str(const char * lang,
 	}
 
 	/* Read dictionary from the input string. */
+
+	/* Make sure "." is used in string conversion to floating point.
+	 * FIXME: Use a locale-agnostic conversion function. */
+	setlocale(LC_NUMERIC, "C");
+
 	dict->input = input;
 	dict->pin = dict->input;
 	if (!read_dictionary(dict))
@@ -290,15 +295,15 @@ Dictionary dictionary_create_from_file(const char * lang)
 }
 
 
-#ifdef USE_VITERBI
+#ifdef NOTDEF
 /**
  * Use "string" as the input dictionary. All of the other parts,
  * including post-processing, affix table, etc, are NULL.
  * This routine is intended for unit-testing ONLY.
  *
- * FIXME? (Used only by code that is in use for now.)
+ * FIXME? (Not used for now.)
  * 1. get_default_locale() returns locale, not language.
- * 2. "lang" memory leak on success.
+ * 2. "lang" memory-leak on success.
  */
 Dictionary dictionary_create_from_utf8(const char * input)
 {
@@ -320,5 +325,5 @@ Dictionary dictionary_create_from_utf8(const char * input)
 
 	return dictionary;
 }
-#endif // USE_VITERBI
+#endif // NOTDEF
 
