@@ -154,13 +154,13 @@ static void dict_error2(Dictionary dict, const char * s, const char *s2)
 
 	if (s2)
 	{
-		prt_error("Error: While parsing dictionary %s:\n"
+		prt_error("Error: While parsing dictionary \"%s\":\n"
 		          "%s %s\n\t Line %d, next tokens: %s\n",
 		          dict->name, s, s2, dict->line_number, tokens);
 	}
 	else
 	{
-		prt_error("Error: While parsing dictionary %s:\n"
+		prt_error("Error: While parsing dictionary \"%s\":\n"
 		          "%s\n\t Line %d, next tokens: %s\n",
 		          dict->name, s, dict->line_number, tokens);
 	}
@@ -349,12 +349,8 @@ static bool link_advance(Dictionary dict)
 				return true;
 			}
 			if (c[0] == 0x0) {
-				if (i == 0) {
-					dict->token[0] = '\0';
-					return true;
-				}
-				dict->token[i] = '\0';
-				dict->already_got_it = '\0';
+				if (i != 0) dict->already_got_it = '\0';
+				dict->token[0] = '\0';
 				return true;
 			}
 			if (lg_isspace(c[0])) {
@@ -403,19 +399,36 @@ static bool check_connector(Dictionary dict, const char * s)
 		return false;
 	}
 	if (*s == '@') s++;
-	if (!isupper((unsigned char)*s) && ('h' != *s) && ('d' != *s)) {
-		dict_error(dict, "The first letter of a connector must be h,d or uppercase.");
+	if (('h' == *s) || ('d' == *s)) s++;
+	if (!isupper((unsigned char)*s)) {
+		dict_error(dict, "Connectors must start with uppercase after an optional h or d.");
 		return false;
 	}
+	/* Note that IDx when x is a subscript is allowed (to allow e.g. ID4id+). */
 	if ((*s == 'I') && (*(s+1) == 'D') && isupper((unsigned char)*(s+2))) {
 		dict_error(dict, "Connectors beginning with \"ID\" are forbidden");
 		return false;
 	}
+
+	bool connector_base = true;
+	s++; /* The first uppercase has been validated above. */
 	while (*(s+1)) {
 		if ((!isalnum((unsigned char)*s)) && (*s != WILD_TYPE)) {
 			dict_error(dict, "All letters of a connector must be ASCII alpha-numeric.");
 			return false;
 		}
+		if (isupper((unsigned char)*s))
+		{
+			if (!connector_base)
+			{
+				dict_error(dict, "Connector subscript contains uppercase.");
+				return false;
+			}
+		}
+		else
+		{
+			connector_base = false;
+		}
 		s++;
 	}
 	return true;
@@ -1251,7 +1264,16 @@ static Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok)
 		 */
 		if (is_number(dict->token))
 		{
-			nl->cost += atof(dict->token);
+			double cost;
+			if (strtodC(dict->token, &cost))
+			{
+				nl->cost += cost;
+			}
+			else
+			{
+				warning(dict, "Invalid cost (using 1.0)\n");
+				nl->cost += 1.0;
+			}
 			if (!link_advance(dict)) {
 				return NULL;
 			}
@@ -1536,9 +1558,9 @@ static void insert_length_limit(Dictionary dict, Dict_node *dn)
 		if ((length_limit < 0) || (length_limit > MAX_SENTENCE) ||
 		  (('\0' != *endp) && (SUBSCRIPT_MARK != *endp)))
 		{
-			prt_error("Warning: Word \"%s\" found near line %d of %s.\n"
+			prt_error("Warning: Word \"%s\" found near line %d of \"%s\".\n"
 					  "\tThis word should end with a number (1-%d).\n"
-					  "\tThis word will be ignored.",
+					  "\tThis word will be ignored.\n",
 					  dn->string, dict->line_number, dict->name, MAX_SENTENCE);
 			return;
 		}
@@ -1588,9 +1610,9 @@ void insert_list(Dictionary dict, Dict_node * p, int l)
 
 	if (is_idiom_word(dn->string))
 	{
-		prt_error("Warning: Word \"%s\" found near line %d of %s.\n"
+		prt_error("Warning: Word \"%s\" found near line %d of \"%s\".\n"
 		        "\tWords ending \".Ix\" (x a number) are reserved for idioms.\n"
-		        "\tThis word will be ignored.",
+		        "\tThis word will be ignored.\n",
 		        dn->string, dict->line_number, dict->name);
 		free(dn);
 	}
@@ -1631,7 +1653,7 @@ void insert_list(Dictionary dict, Dict_node * p, int l)
 						if (!match_found)
 						{
 							prt_error("Warning: The word \"%s\" found near line "
-										 "%d of %s\n\t matches the following words:",
+										 "%d of \"%s\"\n\t matches the following words:",
 										 dn->string, dict->line_number, dict->name);
 							match_found = true;
 						}
@@ -1678,7 +1700,7 @@ static bool read_entry(Dictionary dict)
 			dn = read_word_file(dict, dn, dict->token);
 			if (dn == NULL)
 			{
-				prt_error("Error opening word file %s\n", dict->token);
+				prt_error("Error: Cannot open word file \"%s\".\n", dict->token);
 				return false;
 			}
 		}
