@@ -21,17 +21,17 @@
 #include "fast-match.h"
 #include "resources.h"
 #include "tokenize/word-structures.h" // for Word_struct
+#include "utilities.h"
 
 /* This file contains the exhaustive search algorithm. */
 
 typedef struct Table_connector_s Table_connector;
 struct Table_connector_s
 {
-	Table_connector  *next;
-	Connector        *le, *re;
+	Table_connector  *next;      /* FIXME: eliminate */
+	int              l_id, r_id;
 	Count_bin        count;
-	short            lw, rw;
-	unsigned short   null_count;
+	unsigned short   null_count; /* FIXME: eliminate */
 };
 
 struct count_context_s
@@ -43,8 +43,8 @@ struct count_context_s
 	bool    null_links;
 	bool    exhausted;
 	unsigned int checktimer;  /* Avoid excess system calls */
-	int     table_size;
-	int     log2_table_size;
+	unsigned int table_size;
+	/* int     log2_table_size; */ /* not unused */
 	Table_connector ** table;
 	Resources current_resources;
 };
@@ -78,7 +78,7 @@ static void init_table(count_context_t *ctxt, size_t sent_len)
 	if (24 < shift) shift = 24;
 	lgdebug(+5, "Connector table size (1<<%u)*%zu\n", shift, sizeof(Table_connector));
 	ctxt->table_size = (1U << shift);
-	ctxt->log2_table_size = shift;
+	/* ctxt->log2_table_size = shift; */
 	ctxt->table = (Table_connector**)
 		xalloc(ctxt->table_size * sizeof(Table_connector*));
 	memset(ctxt->table, 0, ctxt->table_size*sizeof(Table_connector*));
@@ -86,9 +86,9 @@ static void init_table(count_context_t *ctxt, size_t sent_len)
 
 }
 
+//#define DEBUG_TABLE_STAT
 #if defined(DEBUG) || defined(DEBUG_TABLE_STAT)
-static int hit;
-static int miss;
+static int hit, miss;  /* Table value found/not found */
 #undef DEBUG_TABLE_STAT
 #define DEBUG_TABLE_STAT(x) x
 /**
@@ -97,40 +97,69 @@ static int miss;
  */
 static void table_stat(count_context_t *ctxt, Sentence sent)
 {
-	int z = 0, nz = 0;
-	int c, N = 0;
-	const int sent_length = sent->length;
-	int *ww = alloca(sent_length*sent_length*sizeof(ww));
-	int *wc = alloca(sent_length*sent_length*sizeof(ww));
+	int z = 0, nz = 0;  /* Number of entries with zero and non-zero counts */
+	int c, total_c = 0; /* Chain length */
+	int N = 0;          /* NULL table slots */
+	int null_count[256] = { 0 };   /* null_count histogram */
+	int chain_length[64] = { 0 };  /* Chain length histogram */
+	int max_chain_length = 0;
 
-	memset(ww, 0, sent_length*sent_length*sizeof(ww));
-	memset(wc, 0, sent_length*sent_length*sizeof(ww));
-	for (int i = 0; i < ctxt->table_size; i++)
+	for (unsigned int i = 0; i < ctxt->table_size; i++)
 	{
 		c = 0;
 		Table_connector *t = ctxt->table[i];
 		if (t == NULL) N++;
 		for (; t != NULL; t = t->next)
 		{
-			if ((t->lw < 0) || (t->rw == sent_length)) continue;
 			c++;
-			if (hist_total(&t->count) == 0) z++;
-			else nz++;
-			ww[t->rw + sent_length * t->lw]++;
-			if (hist_total(&t->count)) wc[t->rw + sent_length * t->lw]++;
+			if (hist_total(&t->count) == 0)
+				z++;
+			else
+				nz++;
+			if (ctxt->null_links > 0)
+				null_count[t->null_count]++;
 		}
-		if (c != 0) printf("Connector table [%d] c=%d\n", i, c);
+		if (c > 0) /* Slot 0 used for length overflow. */
+		{
+			chain_length[c >= (int)ARRAY_SIZE(chain_length) ? 0 : c]++;
+			if (c > max_chain_length) max_chain_length = c;
+			total_c += c;
+		}
+		//if (c != 0) printf("Connector table [%d] c=%d\n", i, c);
 	}
 
-	int wn = 0;
-	for (int i = 0; i < sent_length; i++)
-		for (int j = 0; j < sent_length; j++)
-		{
-			if (ww[i + sent_length * j]) wn++;
-			printf("WW %d %d = %d C=%d\n", i, j, ww[i + sent_length * j], wc[i + sent_length * j]);
-		}
+#if __GNUC__
+#define msb(x) (31-__builtin_clz(x))
+#else
+#define msb(x) 0
+#endif
+	int used_slots = ctxt->table_size-N;
+	/* The used= value is TotalValues/TableSize (not UsedSlots/TableSize). */
+	printf("Connector table: msb=%d slots=%6d/%6d (%5.2f%%) avg-chain=%4.2f "
+	       "values=%6d (z=%5d nz=%5d N=%5d) used=%5.2f%% "
+	       "acc=%d (hit=%d miss=%d) (sent_len=%zu)\n",
+	       msb(ctxt->table_size), used_slots, ctxt->table_size,
+			 100.0f*used_slots/ctxt->table_size, 1.0f*total_c/used_slots,
+	       z+nz, z, nz, N, 100.0f*(z+nz)/ctxt->table_size,
+	       hit+miss, hit, miss, sent->length);
 
-	printf("Connector table z=%d nz=%d N=%d hit=%d miss=%d wc=%d\n", z, nz, N, hit, miss, wn);
+	printf("Chain length histogram:\n");
+	for (int i = 1; i <= max_chain_length; i++)
+		printf("%d: %d\n", i, chain_length[i]);
+	if (chain_length[0] > 0)
+		printf("Chain length > 63: %d\n", chain_length[0]);
+
+	if (ctxt->null_links > 0)
+	{
+		printf("Null count:\n");
+		for (unsigned int i = 0; i < ARRAY_SIZE(null_count); i++)
+		{
+			if (0 != null_count[i])
+				printf("null_count %d: %d\n", i, null_count[i]);
+		}
+	}
+
+	hit = miss = 0;
 }
 #else
 #define DEBUG_TABLE_STAT(x)
@@ -144,13 +173,14 @@ static Table_connector * table_store(count_context_t *ctxt,
                                      Connector *le, Connector *re,
                                      unsigned int null_count)
 {
-	Table_connector *t, *n;
-	unsigned int h;
+	Table_connector *n = pool_alloc(ctxt->sent->Table_connector_pool);
 
-	n = pool_alloc(ctxt->sent->Table_connector_pool);
-	n->lw = lw; n->rw = rw; n->le = le; n->re = re; n->null_count = null_count;
-	h = pair_hash(ctxt->table_size, lw, rw, le, re, null_count);
-	t = ctxt->table[h];
+	unsigned int h = pair_hash(ctxt->table_size, lw, rw, le, re, null_count);
+	Table_connector *t = ctxt->table[h];
+
+	n->l_id = (NULL != le) ? le->suffix_id : lw;
+	n->r_id = (NULL != re) ? re->suffix_id : rw;
+	n->null_count = null_count;
 	n->next = t;
 	ctxt->table[h] = n;
 
@@ -164,13 +194,15 @@ find_table_pointer(count_context_t *ctxt,
                    Connector *le, Connector *re,
                    unsigned int null_count)
 {
-	Table_connector *t;
 	unsigned int h = pair_hash(ctxt->table_size,lw, rw, le, re, null_count);
-	t = ctxt->table[h];
-	for (; t != NULL; t = t->next) {
-		if ((t->lw == lw) && (t->rw == rw)
-		    && (t->le == le) && (t->re == re)
-		    && (t->null_count == null_count))
+	Table_connector *t = ctxt->table[h];
+	int l_id = (NULL != le) ? le->suffix_id : lw;
+	int r_id = (NULL != re) ? re->suffix_id : rw;
+
+	for (; t != NULL; t = t->next)
+	{
+		if ((t->l_id == l_id) && (t->r_id == r_id) &&
+		    (t->null_count == null_count))
 		{
 			DEBUG_TABLE_STAT(hit++);
 			return t;
@@ -260,6 +292,7 @@ static int num_optional_words(count_context_t *ctxt, int w1, int w2)
 	(verbosity_level(D_COUNT_TRACE, "do_count") ? \
 	 prt_error("%-*s", LBLSZ, STRINGIFY(l)) : 0, do_count)
 #define V(c) (!c?"(nil)":connector_string(c))
+#define ID(c,w) (!c?w:c->suffix_id)
 static Count_bin do_count1(int lineno, count_context_t *ctxt,
                           int lw, int rw,
                           Connector *le, Connector *re,
@@ -281,8 +314,8 @@ static Count_bin do_count(int lineno, count_context_t *ctxt,
 		snprintf(m_result, sizeof(m_result), "(M=%lld)", hist_total(&t->count));
 
 	level++;
-	prt_error("%*sdo_count%s:%d lw=%d rw=%d le=%s re=%s null_count=%d\n\\",
-		level*2, "", m_result, lineno, lw, rw, V(le), V(re), null_count);
+	prt_error("%*sdo_count%s:%d lw=%d rw=%d le=%s(%d) re=%s(%d) null_count=%d\n\\",
+		level*2, "", m_result, lineno, lw, rw, V(le),ID(le,lw), V(re),ID(re,rw), null_count);
 	Count_bin r = do_count1(lineno, ctxt, lw, rw, le, re, null_count);
 	prt_error("%*sreturn%.*s:%d=%lld\n",
 	          LBLSZ+level*2, "", (!!t)*3, "(M)", lineno, hist_total(&r));
@@ -697,7 +730,7 @@ void free_count_context(count_context_t *ctxt, Sentence sent)
 {
 	if (NULL == ctxt) return;
 
-	DEBUG_TABLE_STAT(if (verbosity_level(D_SPEC+2)) table_stat(ctxt, sent));
+	DEBUG_TABLE_STAT(if (verbosity_level(+5)) table_stat(ctxt, sent));
 	free_table(ctxt);
 	xfree(ctxt, sizeof(count_context_t));
 }
