@@ -11,18 +11,26 @@
 /*                                                                       */
 /*************************************************************************/
 
-#include <limits.h>  /* For UINT_MAX */
+#include <limits.h>                     // INT_MAX
 
 #include "connectors.h"
 #include "count.h"
-#include "disjunct-utils.h" // for Disjunct
+#include "disjunct-utils.h"             // Disjunct
 #include "extract-links.h"
-#include "utilities.h"                // for Windows rand_r()
+#include "utilities.h"                  // Windows rand_r()
 #include "linkage/linkage.h"
-#include "tokenize/word-structures.h" // for Word_Struct
+#include "tokenize/word-structures.h"   // Word_Struct
 
 //#define RECOUNT
 
+//#define DEBUG_X_TABLE
+#ifdef DEBUG_X_TABLE
+#undef DEBUG_X_TABLE
+#define DEBUG_X_TABLE(...) __VA_ARGS__
+#else
+#define DEBUG_X_TABLE(...)
+#endif
+
 typedef struct Parse_choice_struct Parse_choice;
 
 /* The parse_choice is used to extract links for a given parse */
@@ -31,15 +39,15 @@ struct Parse_choice_struct
 {
 	Parse_choice * next;
 	Parse_set * set[2];
-	Link        link[2];   /* the lc fields of these is NULL if there is no link used */
-	Disjunct *ld, *md, *rd;  /* the chosen disjuncts for the relevant three words */
+	Link        link[2]; /* the lc fields of these is NULL if there is no link used */
+	Disjunct    *md;     /* the chosen disjunct for the middle word */
 };
 
 struct Parse_set_struct
 {
 	short          lw, rw; /* left and right word index */
-	unsigned short null_count; /* number of island words */
-	Connector      *le, *re; /* pending, unconnected connectors */
+	unsigned int   null_count; /* number of island words */
+	int            l_id, r_id; /* pending, unconnected connectors */
 
 	s64 count;      /* The number of ways to parse. */
 #ifdef RECOUNT
@@ -58,7 +66,7 @@ struct Parse_set_struct
 typedef struct Pset_bucket_struct Pset_bucket;
 struct Pset_bucket_struct
 {
-	Parse_set         set;
+	Parse_set set;
 	Pset_bucket *next;
 };
 
@@ -68,6 +76,9 @@ struct extractor_s
 	unsigned int   log2_x_table_size;
 	Pset_bucket ** x_table;  /* Hash table */
 	Parse_set *    parse_set;
+	Word           *words;
+	bool           islands_ok;
+	bool           sort_match_list;
 
 	/* thread-safe random number state */
 	unsigned int rand_state;
@@ -97,7 +108,7 @@ static void free_set(Parse_set *s)
 static Parse_choice *
 make_choice(Parse_set *lset, Connector * llc, Connector * lrc,
             Parse_set *rset, Connector * rlc, Connector * rrc,
-            Disjunct *ld, Disjunct *md, Disjunct *rd)
+            Disjunct *md)
 {
 	Parse_choice *pc;
 	pc = (Parse_choice *) xalloc(sizeof(*pc));
@@ -114,9 +125,7 @@ make_choice(Parse_set *lset, Connector * llc, Connector * lrc,
 	pc->link[1].rw = rset->rw;
 	pc->link[1].lc = rlc;
 	pc->link[1].rc = rrc;
-	pc->ld = ld;
 	pc->md = md;
-	pc->rd = rd;
 	return pc;
 }
 
@@ -141,11 +150,11 @@ static void put_choice_in_set(Parse_set *s, Parse_choice *pc)
 static void record_choice(
     Parse_set *lset, Connector * llc, Connector * lrc,
     Parse_set *rset, Connector * rlc, Connector * rrc,
-    Disjunct *ld, Disjunct *md, Disjunct *rd, Parse_set *s)
+    Disjunct *md, Parse_set *s)
 {
 	put_choice_in_set(s, make_choice(lset, llc, lrc,
 	                                 rset, rlc, rrc,
-	                                 ld, md, rd));
+	                                 md));
 }
 
 /**
@@ -165,17 +174,23 @@ extractor_t * extractor_new(int nwords, unsigned int ranstat)
 	pex->rand_state = ranstat;
 
 	/* Alloc the x_table */
-	if (nwords >= 10) {
-		log2_table_size = 14;
+	if (nwords > 96) {
+		log2_table_size = 15 + nwords / 48;
+	} else if (nwords >= 10) {
+		log2_table_size = 14 + nwords / 24;
 	} else if (nwords >= 4) {
-		log2_table_size = nwords;
+		log2_table_size = 1 + 1.5 * nwords;
 	} else {
-		log2_table_size = 4;
+		log2_table_size = 5;
 	}
+	/* if (log2_table_size > 21) log2_table_size = 21; */
 	pex->log2_x_table_size = log2_table_size;
 	pex->x_table_size = (1 << log2_table_size);
 
-	/*printf("Allocating x_table of size %d\n", x_table_size);*/
+	DEBUG_X_TABLE(
+		printf("Allocating x_table of size %d (nwords %d)\n",
+		       pex->x_table_size, nwords);
+	)
 	pex->x_table = (Pset_bucket**) xalloc(pex->x_table_size * sizeof(Pset_bucket*));
 	memset(pex->x_table, 0, pex->x_table_size * sizeof(Pset_bucket*));
 
@@ -193,18 +208,32 @@ void free_extractor(extractor_t * pex)
 	Pset_bucket *t, *x;
 	if (!pex) return;
 
+	DEBUG_X_TABLE(int N = 0;)
 	for (i=0; i<pex->x_table_size; i++)
 	{
+		DEBUG_X_TABLE(int c = 0;)
 		for (t = pex->x_table[i]; t!= NULL; t=x)
 		{
+			DEBUG_X_TABLE(c++;)
 			x = t->next;
 			free_set(&t->set);
 			xfree((void *) t, sizeof(Pset_bucket));
 		}
+		DEBUG_X_TABLE(
+			if (c > 0)
+				;//printf("I %d: chain %d\n", i, c);
+			else
+				N++;
+		)
 	}
+	DEBUG_X_TABLE(
+		printf("Used x_table %d/%d %.2f%%\n",
+				 pex->x_table_size-N, pex->x_table_size,
+				 100.0f*(pex->x_table_size-N)/pex->x_table_size);
+	)
 	pex->parse_set = NULL;
 
-	/*printf("Freeing x_table of size %d\n", x_table_size);*/
+	//printf("Freeing x_table of size %d\n", pex->x_table_size);
 	xfree((void *) pex->x_table, pex->x_table_size * sizeof(Pset_bucket*));
 	pex->x_table_size = 0;
 	pex->x_table = NULL;
@@ -212,25 +241,59 @@ void free_extractor(extractor_t * pex)
 	xfree((void *) pex, sizeof(extractor_t));
 }
 
+static inline unsigned int el_pair_hash(extractor_t * pex,
+                            int lw, int rw,
+                            const Connector *le, const Connector *re,
+                            unsigned int null_count)
+{
+	unsigned int i;
+	size_t table_size = pex->x_table_size;
+	int l_id = 0, r_id = 0;
+
+	if (NULL != le) l_id = le->suffix_id;
+	if (NULL != re) r_id = re->suffix_id;
+
+#ifdef DEBUG
+	assert(((NULL == le) || le->suffix_id) &&
+	       ((NULL == re) || re->suffix_id));
+#endif
+
+   /* sdbm-based hash */
+	i = null_count;
+	i = lw + (i << 6) + (i << 16) - i;
+	i = rw + (i << 6) + (i << 16) - i;
+	i = l_id + (i << 6) + (i << 16) - i;
+	i = r_id + (i << 6) + (i << 16) - i;
+
+	return i & (table_size-1);
+}
 /**
  * Returns the pointer to this info, NULL if not there.
+ * Note that there is no need to use (lw, rw) as keys because suffix_id
+ * is also encoded by the word number.
  */
 static Pset_bucket * x_table_pointer(int lw, int rw,
                               Connector *le, Connector *re,
                               unsigned int null_count, extractor_t * pex)
 {
 	Pset_bucket *t;
-	t = pex->x_table[pair_hash(pex->x_table_size, lw, rw, le, re, null_count)];
+	t = pex->x_table[el_pair_hash(pex, lw, rw, le, re, null_count)];
+	int l_id = (NULL != le) ? le->suffix_id : lw;
+	int r_id = (NULL != re) ? re->suffix_id : rw;
+
 	for (; t != NULL; t = t->next) {
-		if ((t->set.lw == lw) && (t->set.rw == rw) &&
-		    (t->set.le == le) && (t->set.re == re) &&
-		    (t->set.null_count == null_count))  return t;
+		if ((t->set.l_id == l_id) && (t->set.r_id == r_id) &&
+		    (t->set.null_count == null_count)) return t;
 	}
 	return NULL;
 }
 
 /**
  * Stores the value in the x_table.  Assumes it's not already there.
+ * Since disjuncts of connectors with the same suffix_id may have
+ * different costs and may belong to different words, the disjunct cost
+ * and word_string are saved, to be used as retrieval keys in
+ * x_table_pointer().
  */
 static Pset_bucket * x_table_store(int lw, int rw,
                                   Connector *le, Connector *re,
@@ -243,13 +306,13 @@ static Pset_bucket * x_table_store(int lw, int rw,
 	n->set.lw = lw;
 	n->set.rw = rw;
 	n->set.null_count = null_count;
-	n->set.le = le;
-	n->set.re = re;
+	n->set.l_id = (NULL != le) ? le->suffix_id : lw;
+	n->set.r_id = (NULL != re) ? re->suffix_id : rw;
 	n->set.count = 0;
 	n->set.first = NULL;
 	n->set.tail = NULL;
 
-	h = pair_hash(pex->x_table_size, lw, rw, le, re, null_count);
+	h = el_pair_hash(pex, lw, rw, le, re, null_count);
 	t = pex->x_table[h];
 	n->next = t;
 	pex->x_table[h] = n;
@@ -269,46 +332,33 @@ static Parse_set* dummy_set(int lw, int rw,
 	return &dummy->set;
 }
 
-#ifdef FINISH_THIS_IDEA_MAYBE_LATER
 static int cost_compare(const void *a, const void *b)
 {
-	const Match_node* const * ma = a;
-	const Match_node* const * mb = b;
-	if ((*ma)->d->cost < (*mb)->d->cost) return -1;
-	if ((*ma)->d->cost > (*mb)->d->cost) return 1;
+	const Disjunct * const * da = a;
+	const Disjunct * const * db = b;
+	if ((*da)->cost < (*db)->cost) return -1;
+	if ((*da)->cost > (*db)->cost) return 1;
 	return 0;
 }
 
 /**
- * Sort the matchlist into ascending disjunct cost. The goal here
+ * Sort the match-list into ascending disjunct cost. The goal here
  * is to issue the lowest-cost disjuncts first, so that the parse
  * set ends up quasi-sorted. This is not enough to get us a totally
  * sorted parse set, but it does guarantee that at least the very
  * first parse really will be the lowest cost.
  */
-static Match_node* sort_matchlist(Match_node* mlist)
+static void sort_match_list(fast_matcher_t *mchxt, size_t mlb)
 {
-	Match_node* mx;
-	Match_node** marr;
-	size_t len = 1;
-	size_t i;
+	size_t i = mlb;
 
-	for (mx = mlist; mx->next != NULL; mx = mx->next) len++;
-	if (1 == len) return mlist;
+	while (get_match_list_element(mchxt, i) != NULL)
+		i++;
 
-	/* Avoid blowing out the stack. Its hopeless. */
-	if (100000 < len) return mlist;
+	if (i - mlb < 2) return;
 
-	marr = alloca(len * sizeof(Match_node*));
-	i = 0;
-	for (mx = mlist; mx != NULL; mx = mx->next) marr[i++] = mx;
-
-	qsort((void *) marr, len, sizeof(Match_node*), cost_compare);
-	for (i=0; i<len-1; i++) marr[i]->next = marr[i+1];
-	marr[len-1]->next = NULL;
-	return marr[0];
+	qsort(get_match_list(mchxt, mlb), i - mlb, sizeof(Disjunct *), cost_compare);
 }
-#endif /* FINISH_THIS_IDEA_MAYBE_LATER */
 
 /**
  * returns NULL if there are no ways to parse, or returns a pointer
@@ -321,11 +371,11 @@ static Match_node* sort_matchlist(Match_node* mlist)
  * parse structures.
  */
 static
-Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
+Parse_set * mk_parse_set(fast_matcher_t *mchxt,
                  count_context_t * ctxt,
-                 Disjunct *ld, Disjunct *rd, int lw, int rw,
+                 int lw, int rw,
                  Connector *le, Connector *re, unsigned int null_count,
-                 extractor_t * pex, bool islands_ok)
+                 extractor_t * pex)
 {
 	int start_word, end_word, w;
 	Pset_bucket *xt;
@@ -348,7 +398,7 @@ Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
 	/* This entry must be updated before we return. */
 	xt = x_table_store(lw, rw, le, re, null_count, pex);
 
-	/* The count we previously computed; its non-zero. */
+	/* The count we previously computed; it's non-zero. */
 	xt->set.count = hist_total(count);
 
 #define NUM_PARSES 4
@@ -368,39 +418,39 @@ Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
 		Parse_set* dummy;
 		Disjunct* dis;
 
-		if (!islands_ok && (lw != -1)) return &xt->set;
+		if (!pex->islands_ok && (lw != -1)) return &xt->set;
 		if (null_count == 0) return &xt->set;
 
 		RECOUNT({xt->set.recount = 0;})
 
 		w = lw + 1;
-		for (int opt = 0; opt <= !!words[w].optional; opt++)
+		for (int opt = 0; opt <= !!pex->words[w].optional; opt++)
 		{
 			null_count += opt;
-			for (dis = words[w].d; dis != NULL; dis = dis->next)
+			for (dis = pex->words[w].d; dis != NULL; dis = dis->next)
 			{
 				if (dis->left == NULL)
 				{
-					pset = mk_parse_set(words, mchxt, ctxt,
-											  dis, NULL, w, rw, dis->right, NULL,
-											  null_count-1, pex, islands_ok);
+					pset = mk_parse_set(mchxt, ctxt,
+											  w, rw, dis->right, NULL,
+											  null_count-1, pex);
 					if (pset == NULL) continue;
 					dummy = dummy_set(lw, w, null_count-1, pex);
 					record_choice(dummy, NULL, NULL,
-									  pset,  NULL, NULL,
-									  NULL, NULL, NULL, &xt->set);
+									  pset, dis->right, NULL,
+									  dis, &xt->set);
 					RECOUNT({xt->set.recount += pset->recount;})
 				}
 			}
-			pset = mk_parse_set(words, mchxt, ctxt,
-									  NULL, NULL, w, rw, NULL, NULL,
-									  null_count-1, pex, islands_ok);
+			pset = mk_parse_set(mchxt, ctxt,
+									  w, rw, NULL, NULL,
+									  null_count-1, pex);
 			if (pset != NULL)
 			{
 				dummy = dummy_set(lw, w, null_count-1, pex);
 				record_choice(dummy, NULL, NULL,
 								  pset,  NULL, NULL,
-								  NULL, NULL, NULL, &xt->set);
+								  NULL, &xt->set);
 				RECOUNT({xt->set.recount += pset->recount;})
 			}
 		}
@@ -434,24 +484,24 @@ Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
 	RECOUNT({xt->set.recount = 0;})
 	for (w = start_word; w < end_word; w++)
 	{
-		size_t mlb, mle;
-		mle = mlb = form_match_list(mchxt, w, le, lw, re, rw);
-		// if (mlist) mlist = sort_matchlist(mlist);
-		for (; get_match_list_element(mchxt, mle) != NULL; mle++)
+		size_t mlb = form_match_list(mchxt, w, le, lw, re, rw);
+		if (pex->sort_match_list) sort_match_list(mchxt, mlb);
+
+		for (size_t mle = mlb; get_match_list_element(mchxt, mle) != NULL; mle++)
 		{
-			unsigned int lnull_count, rnull_count;
 			Disjunct *d = get_match_list_element(mchxt, mle);
 			bool Lmatch = d->match_left;
 			bool Rmatch = d->match_right;
 
-			for (lnull_count = 0; lnull_count <= null_count; lnull_count++)
+			for (unsigned int lnull_count = 0; lnull_count <= null_count; lnull_count++)
 			{
 				int i, j;
 				Parse_set *ls[4], *rs[4];
+				bool ls_exists = false;
 
 				/* Here, lnull_count and rnull_count are the null_counts
 				 * we're assigning to those parts respectively. */
-				rnull_count = null_count - lnull_count;
+				unsigned int rnull_count = null_count - lnull_count;
 
 				/* Now, we determine if (based on table only) we can see that
 				   the current range is not parsable. */
@@ -459,46 +509,50 @@ Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
 				for (i=0; i<4; i++) { ls[i] = rs[i] = NULL; }
 				if (Lmatch)
 				{
-					ls[0] = mk_parse_set(words, mchxt, ctxt,
-					             ld, d, lw, w, le->next, d->left->next,
-					             lnull_count, pex, islands_ok);
+					ls[0] = mk_parse_set(mchxt, ctxt,
+					             lw, w, le->next, d->left->next,
+					             lnull_count, pex);
 
 					if (le->multi)
-						ls[1] = mk_parse_set(words, mchxt, ctxt,
-						              ld, d, lw, w, le, d->left->next,
-						              lnull_count, pex, islands_ok);
+						ls[1] = mk_parse_set(mchxt, ctxt,
+						              lw, w, le, d->left->next,
+						              lnull_count, pex);
 
 					if (d->left->multi)
-						ls[2] = mk_parse_set(words, mchxt, ctxt,
-						              ld, d, lw, w, le->next, d->left,
-						              lnull_count, pex, islands_ok);
+						ls[2] = mk_parse_set(mchxt, ctxt,
+						              lw, w, le->next, d->left,
+						              lnull_count, pex);
 
 					if (le->multi && d->left->multi)
-						ls[3] = mk_parse_set(words, mchxt, ctxt,
-						              ld, d, lw, w, le, d->left,
-						              lnull_count, pex, islands_ok);
+						ls[3] = mk_parse_set(mchxt, ctxt,
+						              lw, w, le, d->left,
+						              lnull_count, pex);
+
+					ls_exists =
+						ls[0] != NULL || ls[1] != NULL || ls[2] != NULL || ls[3] != NULL;
 				}
 
-				if (Rmatch)
+
+				if (Rmatch && (ls_exists || le == NULL))
 				{
-					rs[0] = mk_parse_set(words, mchxt, ctxt,
-					                 d, rd, w, rw, d->right->next, re->next,
-					                 rnull_count, pex, islands_ok);
+					rs[0] = mk_parse_set(mchxt, ctxt,
+					                 w, rw, d->right->next, re->next,
+					                 rnull_count, pex);
 
 					if (d->right->multi)
-						rs[1] = mk_parse_set(words, mchxt, ctxt,
-					                 d, rd, w, rw, d->right, re->next,
-						              rnull_count, pex, islands_ok);
+						rs[1] = mk_parse_set(mchxt, ctxt,
+					                 w, rw, d->right, re->next,
+						              rnull_count, pex);
 
 					if (re->multi)
-						rs[2] = mk_parse_set(words, mchxt, ctxt,
-						              d, rd, w, rw, d->right->next, re,
-						              rnull_count, pex, islands_ok);
+						rs[2] = mk_parse_set(mchxt, ctxt,
+						              w, rw, d->right->next, re,
+						              rnull_count, pex);
 
 					if (d->right->multi && re->multi)
-						rs[3] = mk_parse_set(words, mchxt, ctxt,
-						              d, rd, w, rw, d->right, re,
-						              rnull_count, pex, islands_ok);
+						rs[3] = mk_parse_set(mchxt, ctxt,
+						              w, rw, d->right, re,
+						              rnull_count, pex);
 				}
 
 				for (i=0; i<4; i++)
@@ -511,17 +565,17 @@ Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
 						if (rs[j] == NULL) continue;
 						record_choice(ls[i], le, d->left,
 						              rs[j], d->right, re,
-						              ld, d, rd, &xt->set);
+						              d, &xt->set);
 						RECOUNT({xt->set.recount += ls[i]->recount * rs[j]->recount;})
 					}
 				}
 
-				if (ls[0] != NULL || ls[1] != NULL || ls[2] != NULL || ls[3] != NULL)
+				if (ls_exists)
 				{
 					/* Evaluate using the left match, but not the right */
-					Parse_set* rset = mk_parse_set(words, mchxt, ctxt,
-					                        d, rd, w, rw, d->right, re,
-					                        rnull_count, pex, islands_ok);
+					Parse_set* rset = mk_parse_set(mchxt, ctxt,
+					                        w, rw, d->right, re,
+					                        rnull_count, pex);
 					if (rset != NULL)
 					{
 						for (i=0; i<4; i++)
@@ -532,18 +586,18 @@ Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
 							record_choice(ls[i], le, d->left,
 							              rset,  NULL /* d->right */,
 							              re,  /* the NULL indicates no link*/
-							              ld, d, rd, &xt->set);
+							              d, &xt->set);
 							RECOUNT({xt->set.recount += ls[i]->recount * rset->recount;})
 						}
 					}
 				}
-				if ((le == NULL) && (rs[0] != NULL ||
+				else if ((le == NULL) && (rs[0] != NULL ||
 				     rs[1] != NULL || rs[2] != NULL || rs[3] != NULL))
 				{
 					/* Evaluate using the right match, but not the left */
-					Parse_set* lset = mk_parse_set(words, mchxt, ctxt,
-					                        ld, d, lw, w, le, d->left,
-					                        lnull_count, pex, islands_ok);
+					Parse_set* lset = mk_parse_set(mchxt, ctxt,
+					                        lw, w, le, d->left,
+					                        lnull_count, pex);
 
 					if (lset != NULL)
 					{
@@ -553,9 +607,9 @@ Parse_set * mk_parse_set(Word* words, fast_matcher_t *mchxt,
 							/* this ordering is probably not consistent with
 							 * that needed to use list_links */
 							record_choice(lset, NULL /* le */,
-							                    d->left,  /* NULL indicates no link */
+							              d->left,  /* NULL indicates no link */
 							              rs[j], d->right, re,
-							              ld, d, rd, &xt->set);
+							              d, &xt->set);
 							RECOUNT({xt->set.recount += lset->recount * rs[j]->recount;})
 						}
 					}
@@ -621,10 +675,13 @@ bool build_parse_set(extractor_t* pex, Sentence sent,
                     count_context_t *ctxt,
                     unsigned int null_count, Parse_Options opts)
 {
+	pex->words = sent->word;
+	pex->islands_ok = opts->islands_ok;
+	pex->sort_match_list = test_enabled("sort-match-list");
+
 	pex->parse_set =
-		mk_parse_set(sent->word, mchxt, ctxt,
-		             NULL, NULL, -1, sent->length, NULL, NULL, null_count+1,
-		             pex, opts->islands_ok);
+		mk_parse_set(mchxt, ctxt,
+		             -1, sent->length, NULL, NULL, null_count+1, pex);
 
 
 	return set_overflowed(pex);
@@ -640,23 +697,28 @@ void check_link_size(Linkage lkg)
 	}
 }
 
-static void issue_link(Linkage lkg, Disjunct * ld, Disjunct * rd, Link * link)
+/**
+ * Assemble the link array and the chosen_disjuncts of a linkage.
+ */
+static void issue_link(Linkage lkg, bool lr, Disjunct *md, Link *link)
 {
-	check_link_size(lkg);
-	lkg->link_array[lkg->num_links] = *link;
-	lkg->num_links++;
+	if (link->rc != NULL)
+	{
+		check_link_size(lkg);
+		lkg->link_array[lkg->num_links] = *link;
+		lkg->num_links++;
+	}
 
-	lkg->chosen_disjuncts[link->lw] = ld;
-	lkg->chosen_disjuncts[link->rw] = rd;
+	lkg->chosen_disjuncts[lr ? link->lw : link->rw] = md;
 }
 
 static void issue_links_for_choice(Linkage lkg, Parse_choice *pc)
 {
 	if (pc->link[0].lc != NULL) { /* there is a link to generate */
-		issue_link(lkg, pc->ld, pc->md, &pc->link[0]);
+		issue_link(lkg, /*lr*/false, pc->md, &pc->link[0]);
 	}
 	if (pc->link[1].lc != NULL) {
-		issue_link(lkg, pc->md, pc->rd, &pc->link[1]);
+		issue_link(lkg, /*lr*/true, pc->md, &pc->link[1]);
 	}
 }
 
@@ -687,7 +749,6 @@ static void list_random_links(Linkage lkg, extractor_t * pex, const Parse_set *
 	for (pc = set->first; pc != NULL; pc = pc->next) {
 		num_pc++;
 	}
-	assert(num_pc != 0, "Couldn't get a random parse choice");
 
 	new_index = rand_r(&pex->rand_state) % num_pc;
 
