@@ -197,7 +197,6 @@ static Match_node **get_match_table_entry(unsigned int size, Match_node **t,
 			 * unused hash bucket.
 			 */
 			h = (h + 1) & (size-1);
-			if (NULL == t[h]) break;
 			if (h == s) return NULL;
 		}
 	}
@@ -207,7 +206,6 @@ static Match_node **get_match_table_entry(unsigned int size, Match_node **t,
 		{
 			if (con_uc_eq(t[h]->d->left, c)) break;
 			h = (h + 1) & (size-1);
-			if (NULL == t[h]) break;
 			if (h == s) return NULL;
 		}
 	}
@@ -407,7 +405,7 @@ static bool do_match_with_cache(Connector *a, Connector *b, match_cache *c_con)
 	if (c_con->desc == connector_desc(a))
 	{
 		/* The match_cache desc field is initialized to NULL, and this is
-		 * enough because the connector desc filed cannot be NULL, as it
+		 * enough because the connector desc field cannot be NULL, as it
 		 * actually fetched a non-empty match list. */
 		PRAGMA_MAYBE_UNINITIALIZED
 		return c_con->match;
@@ -481,6 +479,21 @@ static bool alt_connection_possible(Connector *c1, Connector *c2,
 #endif /* ALT_CONNECTION_POSSIBLE */
 }
 
+/**
+ * Add match-list termination element (NULL).
+ * Optionally print it (for debug).
+ * Return the match list start.
+ */
+static size_t terminate_match_list(fast_matcher_t *ctxt, int id,
+                             size_t ml_start, int w,
+                             Connector *lc, int lw,
+                             Connector *rc, int rw)
+{
+	push_match_list_element(ctxt, NULL);
+	print_match_list(ctxt, id, ml_start, w, lc, lw, rc, rw);
+	return ml_start;
+}
+
 /**
  * Forms and returns a list of disjuncts coming from word w, that
  * actually matches lc or rc or both. The lw and rw are the words from
@@ -496,6 +509,13 @@ static bool alt_connection_possible(Connector *c1, Connector *c2,
  * disjunct is then added to the result list, and match_right of the
  * same disjunct is set to true when the mr list is processed, and this
  * disjunct is not added again.
+ *
+ * "lc optimization" (see below) marks here a special optimization for the
+ * left connector (lc): If it is non-NULL and doesn't match a disjunct,
+ * there is no need to try to match the right connector (rc), since the
+ * linkage count would be 0 in any case. See do_count() for details. In
+ * that case we can skip the corresponding match list element, or even
+ * immediately return an empty match list if there are no matches for lc.
  */
 size_t
 form_match_list(fast_matcher_t *ctxt, int w,
@@ -503,18 +523,13 @@ form_match_list(fast_matcher_t *ctxt, int w,
                 Connector *rc, int rw)
 {
 	Match_node *mx, *mr_end, **mxp;
-	size_t front = ctxt->match_list_end;
+	size_t front = get_match_list_position(ctxt);
 	Match_node *ml = NULL, *mr = NULL;
 	match_cache mc;
 	gword_cache gc;
 
 	gc.same_alternative = false;
 
-#ifdef VERIFY_MATCH_LIST
-	static int id = 0;
-	int lid = ++id; /* A local copy, for multi-threading support. */
-#endif
-
 	/* Get the lists of candidate matching disjuncts of word w for lc and
 	 * rc.  Consider each of these lists only if the length_limit of lc
 	 * rc and also w, is not greater then the distance between their word
@@ -524,12 +539,23 @@ form_match_list(fast_matcher_t *ctxt, int w,
 		mxp = get_match_table_entry(ctxt->l_table_size[w], ctxt->l_table[w], lc, -1);
 		if (NULL != mxp) ml = *mxp;
 	}
+	if ((lc != NULL) && (ml == NULL)) /* lc optimization */
+		return terminate_match_list(ctxt, -1, front, w, lc, lw, rc, rw);
+
 	if ((rc != NULL) && ((rw - w) <= rc->length_limit))
 	{
 		mxp = get_match_table_entry(ctxt->r_table_size[w], ctxt->r_table[w], rc, 1);
 		if (NULL != mxp) mr = *mxp;
 	}
 
+	if ((ml == NULL) && (mr == NULL))
+		return terminate_match_list(ctxt, -2, front, w, lc, lw, rc, rw);
+
+#ifdef VERIFY_MATCH_LIST
+	static int id = 0;
+	int lid = ++id; /* A local copy, for multi-threading support. */
+#endif
+
 	for (mx = mr; mx != NULL; mx = mx->next)
 	{
 		if (mx->d->right->nearest_word > rw) break;
@@ -556,6 +582,9 @@ form_match_list(fast_matcher_t *ctxt, int w,
 		push_match_list_element(ctxt, mx->d);
 	}
 
+	if ((lc != NULL) && is_no_match_list(ctxt, front)) /* lc optimization */
+		return terminate_match_list(ctxt, -3, front, w, lc, lw, rc, rw);
+
 	/* Append the list of things that could match the right.
 	 * Note that it is important to set here match_right correctly even
 	 * if we are going to skip this element here because its match_left
@@ -567,6 +596,7 @@ form_match_list(fast_matcher_t *ctxt, int w,
 	{
 		if ((rw - w) > mx->d->right->length_limit) continue;
 
+		if ((lc != NULL) && !mx->d->match_left) continue; /* lc optimization */
 		mx->d->match_right = do_match_with_cache(mx->d->right, rc, &mc) &&
 			                  alt_connection_possible(mx->d->right, rc, &gc);
 		if (!mx->d->match_right || mx->d->match_left) continue;
@@ -577,7 +607,5 @@ form_match_list(fast_matcher_t *ctxt, int w,
 		push_match_list_element(ctxt, mx->d);
 	}
 
-	push_match_list_element(ctxt, NULL);
-	print_match_list(ctxt, lid, front, w, lc, lw, rc, rw);
-	return front;
+	return terminate_match_list(ctxt, lid, front, w, lc, lw, rc, rw);
 }
