@@ -24,9 +24,9 @@
 #include <sys/types.h>
 #include <stdarg.h>
 #include <locale.h>
-#ifdef HAVE_LOCALE_T_IN_XLOCALE_H
+#ifdef HAVE_XLOCALE_H
 #include <xlocale.h>
-#endif /* HAVE_LOCALE_T_IN_XLOCALE_H */
+#endif /* HAVE_XLOCALE_H */
 
 #ifndef _WIN32
 	// #include <unistd.h>
@@ -658,6 +658,49 @@ char * get_default_locale(void)
 	return safe_strdup(locale);
 }
 
+#ifdef HAVE_LOCALE_T
+static void free_C_LC_NUMERIC(void);
+
+static locale_t get_C_LC_NUMERIC(void)
+{
+	static locale_t locobj;
+
+	if ((locale_t)0 != locobj) return locobj;
+
+#ifdef _WIN32
+	locobj = _create_locale(LC_NUMERIC, "C");
+#else
+	locobj = newlocale(LC_NUMERIC_MASK, "C", (locale_t)0);
+#endif /* _WIN32 */
+
+	atexit(free_C_LC_NUMERIC);
+
+	return locobj;
+}
+
+static void free_C_LC_NUMERIC(void)
+{
+	freelocale(get_C_LC_NUMERIC());
+}
+#endif /* HAVE_LOCALE_T */
+
+bool strtodC(const char *s, double *r)
+{
+	char *err;
+
+#ifdef HAVE_LOCALE_T
+	double val = strtod_l(s, &err, get_C_LC_NUMERIC());
+#else
+	/* dictionary_setup_locale() invokes setlocale(LC_NUMERIC, "C") */
+	double val = strtod(s, &err);
+#endif /* HAVE_LOCALE_T */
+
+	if ('\0' != *err) return false; /* *r unaffected */
+
+	*r = val;
+	return true;
+}
+
 /* ============================================================= */
 /* Alternatives utilities */
 
